#include "stdafx.h"
#include "Deferred.h"
#include "Matrix.h"
#define MAX_LOADSTRING 128

#include <d3dx9.h>

#include <cstdio>

HINSTANCE hInst;
TCHAR szTitle[MAX_LOADSTRING];
TCHAR szWindowClass[MAX_LOADSTRING];

ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About(HWND, UINT, WPARAM, LPARAM);

/******************************************************************************\
\******************************************************************************/

#define PI 3.14159265359f

const float _nearClip = 1.0f / 16.0f;

const int _xRes = 800;
const int _yRes = 600;

static IDirect3D9*			pD3D;
static IDirect3DDevice9*	pDevice;

static Vec3		_cubePositions[] =
{
	Vec3(	-1.0f,	-1.0f,	-1.0f	),
	Vec3(	-1.0f,	 1.0f,	-1.0f	),
	Vec3(	 1.0f,	-1.0f,	-1.0f	),
	Vec3(	 1.0f,	 1.0f,	-1.0f	),

	Vec3(	-1.0f,	-1.0f,	 1.0f	),
	Vec3(	 1.0f,	-1.0f,	 1.0f	),
	Vec3(	-1.0f,	 1.0f,	 1.0f	),
	Vec3(	 1.0f,	 1.0f,	 1.0f	),

	Vec3(	-1.0f,	-1.0f,	-1.0f	),
	Vec3(	 1.0f,	-1.0f,	-1.0f	),
	Vec3(	-1.0f,	-1.0f,	 1.0f	),
	Vec3(	 1.0f,	-1.0f,	 1.0f	),

	Vec3(	-1.0f,	 1.0f,	-1.0f	),
	Vec3(	-1.0f,	 1.0f,	 1.0f	),
	Vec3(	 1.0f,	 1.0f,	-1.0f	),
	Vec3(	 1.0f,	 1.0f,	 1.0f	),

	Vec3(	-1.0f,	-1.0f,	-1.0f	),
	Vec3(	-1.0f,	-1.0f,	 1.0f	),
	Vec3(	-1.0f,	 1.0f,	-1.0f	),
	Vec3(	-1.0f,	 1.0f,	 1.0f	),

	Vec3(	 1.0f,	-1.0f,	-1.0f	),
	Vec3(	 1.0f,	 1.0f,	-1.0f	),
	Vec3(	 1.0f,	-1.0f,	 1.0f	),
	Vec3(	 1.0f,	 1.0f,	 1.0f	),
};

static Vec3		_cubeNorms[] =
{
	Vec3(	  0.0f,	 0.0f,	-1.0f	),
	Vec3(	  0.0f,	 0.0f,	-1.0f	),
	Vec3(	  0.0f,	 0.0f,	-1.0f	),
	Vec3(	  0.0f,	 0.0f,	-1.0f	),

	Vec3(	  0.0f,	 0.0f,	 1.0f	),
	Vec3(	  0.0f,	 0.0f,	 1.0f	),
	Vec3(	  0.0f,	 0.0f,	 1.0f	),
	Vec3(	  0.0f,	 0.0f,	 1.0f	),

	Vec3(	  0.0f,	-1.0f,	 0.0f	),
	Vec3(	  0.0f,	-1.0f,	 0.0f	),
	Vec3(	  0.0f,	-1.0f,	 0.0f	),
	Vec3(	  0.0f,	-1.0f,	 0.0f	),

	Vec3(	  0.0f,	 1.0f,	 0.0f	),
	Vec3(	  0.0f,	 1.0f,	 0.0f	),
	Vec3(	  0.0f,	 1.0f,	 0.0f	),
	Vec3(	  0.0f,	 1.0f,	 0.0f	),

	Vec3(	 -1.0f,	 0.0f,	 0.0f	),
	Vec3(	 -1.0f,	 0.0f,	 0.0f	),
	Vec3(	 -1.0f,	 0.0f,	 0.0f	),
	Vec3(	 -1.0f,	 0.0f,	 0.0f	),

	Vec3(	  1.0f,	 0.0f,	 0.0f	),
	Vec3(	  1.0f,	 0.0f,	 0.0f	),
	Vec3(	  1.0f,	 0.0f,	 0.0f	),
	Vec3(	  1.0f,	 0.0f,	 0.0f	),
};

static Vec2		_cubeTexCoords[] =
{
	Vec2(	  0.0f,	 0.0f	),
	Vec2(	  0.0f,	 1.0f	),
	Vec2(	  1.0f,	 0.0f	),
	Vec2(	  1.0f,	 1.0f	),

	Vec2(	  0.0f,	 0.0f	),
	Vec2(	  0.0f,	 1.0f	),
	Vec2(	  1.0f,	 0.0f	),
	Vec2(	  1.0f,	 1.0f	),

	Vec2(	  0.0f,	 0.0f	),
	Vec2(	  0.0f,	 1.0f	),
	Vec2(	  1.0f,	 0.0f	),
	Vec2(	  1.0f,	 1.0f	),

	Vec2(	  0.0f,	 0.0f	),
	Vec2(	  0.0f,	 1.0f	),
	Vec2(	  1.0f,	 0.0f	),
	Vec2(	  1.0f,	 1.0f	),

	Vec2(	  0.0f,	 0.0f	),
	Vec2(	  0.0f,	 1.0f	),
	Vec2(	  1.0f,	 0.0f	),
	Vec2(	  1.0f,	 1.0f	),

	Vec2(	  0.0f,	 0.0f	),
	Vec2(	  0.0f,	 1.0f	),
	Vec2(	  1.0f,	 0.0f	),
	Vec2(	  1.0f,	 1.0f	),
};

typedef int Tri[3];

static Tri		_cubeFaces[] =
{
	{	0,	1,	2	},
	{	3,	2,	1	},

	{	4,	5,	6	},
	{	7,	6,	5	},

	{	8,	9,	10	},
	{	11,	10,	9	},

	{	12,	13,	14	},
	{	15,	14,	13	},

	{	16,	17,	18	},
	{	19,	18,	17	},

	{	20,	21,	22	},
	{	23,	22,	21	},
};

static const int	_numCubeVerts = sizeof(_cubePositions) / sizeof(*_cubePositions);
static const int	_numCubeFaces = sizeof(_cubeFaces) / sizeof(*_cubeFaces);

class Light
{
public:
	Light()	{}
	Light(const Vec3& color): _color(color)	{}
	
	void		SetColor(const Vec3& color)	{ _color = color; }
	const Vec3&	GetColor()					{ return _color; }
	
	virtual void Render(const Mat44& view, const Mat44& proj) = 0;

protected:
	Vec3	_color;
};

struct SunLight: public Light
{
public:
	SunLight(): Light()	{}
	SunLight(const Vec3& color, const Vec3& dir): Light(color), _dir(dir)	{}

	void		SetDirection(const Vec3& dir)	{ _dir = dir; }
	const Vec3&	GetDirection()					{ return _dir; }

	virtual void Render(const Mat44& view, const Mat44& proj);

protected:
	Vec3	_dir;
};

struct OmniLight: public Light
{
public:
	OmniLight(): Light()	{}
	OmniLight(const Vec3& color, const Vec3& pos, float rad): Light(color), _pos(pos), _rad(rad)	{}

	void		SetPosition(const Vec3& pos)	{ _pos = pos; }
	const Vec3&	GetPosition()					{ return _pos; }

	void		SetRadius(float rad)			{ _rad = rad; }
	float		GetRadius()						{ return _rad; }

	virtual void Render(const Mat44& view, const Mat44& proj);

protected:
	Vec3	_pos;
	float	_rad;
};

static OmniLight	_omnis[] =
{
	OmniLight(	Vec3(2.0f, 1.0f, 0.0f),	Vec3(-2.5f, 0.0f, 0.0f),	4.0f),
	OmniLight(	Vec3(0.0f, 2.0f, 1.0f),	Vec3(2.5f, 0.0f, 0.0f),		3.0f),
	OmniLight(	Vec3(1.0f, 0.0f, 2.0f),	Vec3(0.0f, -2.5f, 0.0f),	3.0f),
	OmniLight(	Vec3(2.0f, 0.0f, 1.0f),	Vec3(0.0f, 2.5f, 0.0f),		4.0f),
};
static int _numOmnis = sizeof(_omnis) / sizeof(*_omnis);

static IDirect3DVertexDeclaration9*	_pVertDecl;

static IDirect3DVertexBuffer9*	_pCubePosBuffer;
static IDirect3DVertexBuffer9*	_pCubeNormBuffer;
static IDirect3DVertexBuffer9*	_pCubeTexCoordBuffer;
static IDirect3DIndexBuffer9*	_pCubeFaceBuffer;

static IDirect3DTexture9*		_pNormTargTex;
static IDirect3DTexture9*		_pRefTargTex;
static IDirect3DTexture9*		_pDepthTargTex;
static IDirect3DTexture9*		_pDiffTargTex;

static IDirect3DVertexShader9*	_pGeomVertShader;
static IDirect3DPixelShader9*	_pGeomPixShader;
static IDirect3DVertexShader9*	_pLightVertShader;
static IDirect3DPixelShader9*	_pOmniLightPixShader;
static IDirect3DPixelShader9*	_pAmbientPixShader;

static IDirect3DTexture9*		_pDiffuseTex;
static IDirect3DTexture9*		_pSpecularTex;

static 	D3DXMATRIX	_identityMat(	1.0f, 0.0f, 0.0f, 0.0f,
									0.0f, 1.0f, 0.0f, 0.0f,
									0.0f, 0.0f, 1.0f, 0.0f,
									0.0f, 0.0f, 0.0f, 1.0f);

void D3DErr(HRESULT d3dErr)
{
	switch(d3dErr)
	{
		case D3D_OK:
			printf("D3D_OK\n");
			break;
		case D3DERR_WRONGTEXTUREFORMAT:
			printf("D3DERR_WRONGTEXTUREFORMAT\n");
			break;
		case D3DERR_UNSUPPORTEDCOLOROPERATION:
			printf("D3DERR_UNSUPPORTEDCOLOROPERATION\n");
			break;
		case D3DERR_UNSUPPORTEDCOLORARG:
			printf("D3DERR_UNSUPPORTEDCOLORARG\n");
			break;
		case D3DERR_UNSUPPORTEDALPHAOPERATION:
			printf("D3DERR_UNSUPPORTEDALPHAOPERATION\n");
			break;
		case D3DERR_UNSUPPORTEDALPHAARG:
			printf("D3DERR_UNSUPPORTEDALPHAARG\n");
			break;
		case D3DERR_TOOMANYOPERATIONS:
			printf("D3DERR_TOOMANYOPERATIONS\n");
			break;
		case D3DERR_CONFLICTINGTEXTUREFILTER:
			printf("D3DERR_CONFLICTINGTEXTUREFILTER\n");
			break;
		case D3DERR_UNSUPPORTEDFACTORVALUE:
			printf("D3DERR_UNSUPPORTEDFACTORVALUE\n");
			break;
		case D3DERR_CONFLICTINGRENDERSTATE:
			printf("D3DERR_CONFLICTINGRENDERSTATE\n");
			break;
		case D3DERR_UNSUPPORTEDTEXTUREFILTER:
			printf("D3DERR_UNSUPPORTEDTEXTUREFILTER\n");
			break;
		case D3DERR_CONFLICTINGTEXTUREPALETTE:
			printf("D3DERR_CONFLICTINGTEXTUREPALETTE\n");
			break;
		case D3DERR_DRIVERINTERNALERROR:
			printf("D3DERR_DRIVERINTERNALERROR\n");
			break;
		case D3DERR_NOTFOUND:
			printf("D3DERR_NOTFOUND\n");
			break;
		case D3DERR_MOREDATA:
			printf("D3DERR_MOREDATA\n");
			break;
		case D3DERR_DEVICELOST:
			printf("D3DERR_DEVICELOST\n");
			break;
		case D3DERR_DEVICENOTRESET:
			printf("D3DERR_DEVICENOTRESET\n");
			break;
		case D3DERR_NOTAVAILABLE:
			printf("D3DERR_NOTAVAILABLE\n");
			break;
		case D3DERR_OUTOFVIDEOMEMORY:
			printf("D3DERR_OUTOFVIDEOMEMORY\n");
			break;
		case D3DERR_INVALIDDEVICE:
			printf("D3DERR_INVALIDDEVICE\n");
			break;
		case D3DERR_INVALIDCALL:
			printf("D3DERR_INVALIDCALL\n");
			break;
		case D3DERR_DRIVERINVALIDCALL:
			printf("D3DERR_DRIVERINVALIDCALL\n");
			break;
		case D3DERR_WASSTILLDRAWING:
			printf("D3DERR_WASSTILLDRAWING\n");
			break;
		case D3DOK_NOAUTOGEN:
			printf("D3DOK_NOAUTOGEN\n");
			break;
		default:
			printf("Unknown D3D error.\n");
			break;
	}
}

Mat44 FlattenProjection(const Mat44& proj, float depth)
{
	Mat44 result = proj;

	result.Set(2, 2) = proj.Get(2, 2) + proj.Get(3, 2) / max(depth, abs(proj.Get(3, 2)));
	result.Set(3, 2) = 0.0f;
	
	return result;
}

void SunLight::Render(const Mat44& view, const Mat44& proj)
{
}

void OmniLight::Render(const Mat44& view, const Mat44& proj)
{
	D3DXVECTOR4	unprojScale(1.0f / proj.Get(0, 0), 1.0f / proj.Get(1, 1), 1.0f, 1.0f);

	Vec4 pos = view * Vec4(_pos, 1.0f);
	Vec4 color = Vec4(_color, 1.0f / (_rad * _rad));

	Mat44 flatProj = FlattenProjection(proj, pos.Get(2) - _rad);

	Mat44 world = Mat44::Identity() * _rad;
	world.Set(3) = Vec4(_pos, 1.0f);
	
	D3DXMATRIX vp = *(D3DXMATRIX*)&view * *(D3DXMATRIX*)&flatProj;
	D3DXMATRIX wvp = *(D3DXMATRIX*)&world * *(D3DXMATRIX*)&vp;

	D3DXMatrixTranspose(&wvp, &wvp);

	pDevice->SetVertexShaderConstantF(0, (float*)&wvp, 4);
	pDevice->SetVertexShaderConstantF(8, (float*)&unprojScale, 1);

	pDevice->SetPixelShaderConstantF(1, (float*)&pos, 1);
	pDevice->SetPixelShaderConstantF(2, (float*)&color, 1);

	pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, _numCubeVerts, 0, _numCubeFaces);
}

void RenderScene(float t, const D3DXMATRIX& proj, const D3DXMATRIX& view)
{
	D3DXMATRIX temp;

	pDevice->SetIndices(_pCubeFaceBuffer);

	pDevice->SetStreamSource(	0,	_pCubePosBuffer,		0,	sizeof(*_cubePositions)	);
	pDevice->SetStreamSource(	1,	_pCubeNormBuffer,		0,	sizeof(*_cubeNorms)		);
	pDevice->SetStreamSource(	2,	_pCubeTexCoordBuffer,	0,	sizeof(*_cubeTexCoords)	);

	pDevice->SetTexture(0, _pDiffuseTex);
	pDevice->SetTexture(1, _pSpecularTex);

	D3DXMATRIX vp = view * proj;

	D3DXMATRIX	world;
	D3DXMatrixRotationYawPitchRoll(&world, -t * PI * 1.0f, 0.0f, t * PI * 0.5f);

	D3DXMATRIX wv = world * view;
	D3DXMATRIX wvp = world * vp;
	
	D3DXMatrixTranspose(&wv, &wv);
	D3DXMatrixTranspose(&wvp, &wvp);

	pDevice->SetVertexShaderConstantF(0, (float*)&wvp, 4);
	pDevice->SetVertexShaderConstantF(4, (float*)&wv, 4);

	D3DXVECTOR4 color(1.0f, 0.0f, 1.0f, 1.0f);
	D3DXVECTOR4 material(1.0f, 1.0f, 1.0f, 1.0f);		// (Unused, Self Illum, Gloss, Specular)

	pDevice->SetPixelShaderConstantF(0, (float*)&color, 1);
	pDevice->SetPixelShaderConstantF(1, (float*)&material, 1);

	pDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, _numCubeVerts, 0, _numCubeFaces);


	D3DXMatrixRotationYawPitchRoll(&world, -t * PI * 0.5f, -t * PI, 0.0f );

	//world = _identityMat;

	wv = -1.0f * world * view;		// -1.0f here will effectively flip the vertex normals.
	world *= 4.0f;
	world(3, 3) = 1.0f;
	wvp = world * vp;
	
	D3DXMatrixTranspose(&wv, &wv);
	D3DXMatrixTranspose(&wvp, &wvp);

	pDevice->SetVertexShaderConstantF(0, (float*)&wvp, 4);
	pDevice->SetVertexShaderConstantF(4, (float*)&wv, 4);

	color = D3DXVECTOR4(1.0f, 1.0f, 0.0f, 1.0f);
	material = D3DXVECTOR4(1.0f, 0.0f, 1.0f, 1.0f);		// (Unused, Self Illum, Gloss, Specular)

	pDevice->SetPixelShaderConstantF(0, (float*)&color, 1);
	pDevice->SetPixelShaderConstantF(1, (float*)&material, 1);

	pDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

	pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, _numCubeVerts, 0, _numCubeFaces);
}

void Paint(void)
{
	static float t = 0.0f;

	if (pDevice == 0) 
		return;

	// All the good stuff goes here.

	pDevice->BeginScene();

	pDevice->Clear(
		0,
		NULL,
		D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
		0,
		1.0f,
		0);

	IDirect3DSurface9*	pBackSurf;
	pDevice->GetRenderTarget(0, &pBackSurf);
	
	{
		IDirect3DSurface9*	pRenderSurf;

		_pNormTargTex->GetSurfaceLevel(0, &pRenderSurf);
		pDevice->SetRenderTarget(0, pRenderSurf);

		_pRefTargTex->GetSurfaceLevel(0, &pRenderSurf);
		pDevice->SetRenderTarget(1, pRenderSurf);

		_pDiffTargTex->GetSurfaceLevel(0, &pRenderSurf);
		pDevice->SetRenderTarget(2, pRenderSurf);

		_pDepthTargTex->GetSurfaceLevel(0, &pRenderSurf);
		pDevice->SetRenderTarget(3, pRenderSurf);
	}

	D3DXMATRIX	proj;
	D3DXMatrixPerspectiveFovLH(&proj, PI/2.2f, 4.0f/3.0f, 0.0f, 0.0f);

	D3DXVECTOR4	unprojScale(1.0f / proj(0, 0), 1.0f / proj(1, 1), 1.0f, 1.0f);

	// Finish the perpecive matrix setup here, using an infinitely far clipping plane (in effect, disabling far clipping).
	proj(2, 2) = 1.0f;
	proj(3, 2) = -_nearClip;

	D3DXMATRIX	view;
	D3DXMatrixTranslation(&view, 0.0f, 0.0f, 5.0f);

	pDevice->SetVertexShader(_pGeomVertShader);
	pDevice->SetPixelShader(_pGeomPixShader);

	pDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	pDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
	
	pDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetSamplerState(3, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(3, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	pDevice->SetSamplerState(3, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetVertexShaderConstantF(8, (float*)&unprojScale, 1);

	// Render all scene geometry
	RenderScene(t, proj, view);

	// Lighting
	pDevice->SetRenderTarget(0, pBackSurf);
	pDevice->SetRenderTarget(1, NULL);
	pDevice->SetRenderTarget(2, NULL);
	pDevice->SetRenderTarget(3, NULL);

	pDevice->SetTexture(0, _pNormTargTex);
	pDevice->SetTexture(1, _pRefTargTex);
	pDevice->SetTexture(2, _pDiffTargTex);
	pDevice->SetTexture(3, _pDepthTargTex);

	pDevice->SetVertexShader(_pLightVertShader);
	pDevice->SetPixelShader(_pOmniLightPixShader);

	pDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
	pDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
	pDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	pDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
	pDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

	// Point sampling is all that's necessary here, and so cheap...
	pDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	pDevice->SetSamplerState(3, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(3, D3DSAMP_MINFILTER, D3DTEXF_POINT);
	pDevice->SetSamplerState(3, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

	for (int i = 0; i < _numOmnis; i++)
	{
		_omnis[i].Render(*(Mat44*)&view, *(Mat44*)&proj);
	}

	// Ambient/emissive pass
	pDevice->SetPixelShader(_pAmbientPixShader);

	D3DXMatrixOrthoOffCenterLH(&proj, -1.0f, 1.0f, -1.0f, 1.0f, -1.0f, 1.0f);
	D3DXMatrixTranspose(&proj, &proj);
	pDevice->SetVertexShaderConstantF(0, (float*)&proj, 4);
	
	Vec4 ambient(0.125f, 0.125f, 0.125f, 1.0f);
	pDevice->SetPixelShaderConstantF(2, (float*)&ambient, 1);

	pDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

	pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, 4, 0, 2);



	pDevice->EndScene();

	pDevice->Present( NULL, NULL, NULL, NULL );
	
	t += 1.0f / 128.0f;
}

int APIENTRY _tWinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
	MSG msg;
	HACCEL hAccelTable;

	// Initialize global strings
	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
	LoadString(hInstance, IDC_DEFERRED, szWindowClass, MAX_LOADSTRING);
	MyRegisterClass(hInstance);

	// Perform application initialization:
	if (!InitInstance (hInstance, nCmdShow)) 
	{
		return FALSE;
	}

	hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_DEFERRED);

	// Main message loop:
	int keepGoing = 1;

	while (keepGoing)
	{
		while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			keepGoing = GetMessage(&msg, NULL, 0, 0);
			if (keepGoing)
			{
				if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
		}
		Paint();
	}

	return (int) msg.wParam;
}

//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    This function and its usage are only necessary if you want this code
//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
//    function that was added to Windows 95. It is important to call this function
//    so that the application will get 'well formed' small icons associated
//    with it.

ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_DEFERRED);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= (LPCTSTR)IDC_DEFERRED;
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

	return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HANDLE, int)
//
//   PURPOSE: Saves instance handle and creates main window
//
//   COMMENTS:
//
//        In this function, we save the instance handle in a global variable and
//        create and display the main program window.

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	HWND hWnd;

	hInst = hInstance; // Store instance handle in our global variable

	hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

	if (!hWnd)
	{
		return FALSE;
	}

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	pD3D = Direct3DCreate9( D3D_SDK_VERSION );

	D3DPRESENT_PARAMETERS	presentParams;
	ZeroMemory( &presentParams, sizeof(presentParams) );
	presentParams.Windowed = TRUE;//FALSE;
	presentParams.SwapEffect = D3DSWAPEFFECT_DISCARD;
	presentParams.BackBufferFormat = D3DFMT_A8R8G8B8;
	presentParams.EnableAutoDepthStencil = TRUE;
	presentParams.AutoDepthStencilFormat = D3DFMT_D24S8;
	presentParams.BackBufferWidth = _xRes;
	presentParams.BackBufferHeight = _yRes;

	pD3D->CreateDevice  
	( 
		D3DADAPTER_DEFAULT, 
		D3DDEVTYPE_HAL, 
		hWnd, 
		D3DCREATE_HARDWARE_VERTEXPROCESSING, 
		&presentParams, 
		&pDevice
	);

	D3DVERTEXELEMENT9 vertDecl[] =
	{
		{ 0, 0,		D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,	0 },
		{ 1, 0,		D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,		0 },
		{ 2, 0,		D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	0 },
		D3DDECL_END()
	};

	pDevice->CreateVertexDeclaration(vertDecl, &_pVertDecl);
	pDevice->SetVertexDeclaration(_pVertDecl);

	pDevice->CreateVertexBuffer(	sizeof(_cubePositions),	0, 0,				D3DPOOL_DEFAULT, &_pCubePosBuffer,		NULL	);
	pDevice->CreateVertexBuffer(	sizeof(_cubeNorms),		0, 0,				D3DPOOL_DEFAULT, &_pCubeNormBuffer,		NULL	);
	pDevice->CreateVertexBuffer(	sizeof(_cubeTexCoords),	0, 0,				D3DPOOL_DEFAULT, &_pCubeTexCoordBuffer,	NULL	);
	pDevice->CreateIndexBuffer(		sizeof(_cubeFaces),		0, D3DFMT_INDEX32,	D3DPOOL_DEFAULT, &_pCubeFaceBuffer,		NULL	);

	void*	pData;

	_pCubePosBuffer->Lock(0, 0, &pData, 0);
	memcpy(pData, _cubePositions, sizeof(_cubePositions));
	_pCubePosBuffer->Unlock();
	
	_pCubeNormBuffer->Lock(0, 0, &pData, 0);
	memcpy(pData, _cubeNorms, sizeof(_cubeNorms));
	_pCubeNormBuffer->Unlock();

	_pCubeTexCoordBuffer->Lock(0, 0, &pData, 0);
	memcpy(pData, _cubeTexCoords, sizeof(_cubeTexCoords));
	_pCubeTexCoordBuffer->Unlock();

	_pCubeFaceBuffer->Lock(0, 0, &pData, 0);
	memcpy(pData, _cubeFaces, sizeof(_cubeFaces));
	_pCubeFaceBuffer->Unlock();

	pDevice->CreateTexture(	_xRes,	_yRes,	1,	D3DUSAGE_RENDERTARGET,	D3DFMT_A8R8G8B8,	D3DPOOL_DEFAULT,	&_pNormTargTex,		NULL	);
	pDevice->CreateTexture(	_xRes,	_yRes,	1,	D3DUSAGE_RENDERTARGET,	D3DFMT_A8R8G8B8,	D3DPOOL_DEFAULT,	&_pRefTargTex,		NULL	);
	pDevice->CreateTexture(	_xRes,	_yRes,	1,	D3DUSAGE_RENDERTARGET,	D3DFMT_A8R8G8B8,	D3DPOOL_DEFAULT,	&_pDiffTargTex,		NULL	);
	pDevice->CreateTexture(	_xRes,	_yRes,	1,	D3DUSAGE_RENDERTARGET,	D3DFMT_R32F,		D3DPOOL_DEFAULT,	&_pDepthTargTex,	NULL	);

	pDevice->SetRenderState(D3DRS_LIGHTING, FALSE);

	ID3DXBuffer* pCode;

	D3DXAssembleShaderFromFile("Preshade.vsh", NULL, NULL, 0, &pCode, NULL);
	pDevice->CreateVertexShader((DWORD*)pCode->GetBufferPointer(), &_pGeomVertShader);

	D3DXAssembleShaderFromFile("Preshade.psh", NULL, NULL, 0, &pCode, NULL);
	pDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), &_pGeomPixShader);

	D3DXAssembleShaderFromFile("Light.vsh", NULL, NULL, 0, &pCode, NULL);
	pDevice->CreateVertexShader((DWORD*)pCode->GetBufferPointer(), &_pLightVertShader);

	D3DXAssembleShaderFromFile("OmniLight.psh", NULL, NULL, 0, &pCode, NULL);
	pDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), &_pOmniLightPixShader);

	D3DXAssembleShaderFromFile("Ambient.psh", NULL, NULL, 0, &pCode, NULL);
	pDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), &_pAmbientPixShader);

	D3DXCreateTextureFromFile(pDevice, "diffuse.dds", &_pDiffuseTex);
	D3DXCreateTextureFromFile(pDevice, "specular.dds", &_pSpecularTex);

	pDevice->SetSamplerState(0, D3DSAMP_MAXANISOTROPY, 8);
	pDevice->SetSamplerState(1, D3DSAMP_MAXANISOTROPY, 8);
	pDevice->SetSamplerState(2, D3DSAMP_MAXANISOTROPY, 8);
	pDevice->SetSamplerState(3, D3DSAMP_MAXANISOTROPY, 8);

	return TRUE;
}

//
//  FUNCTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND	- process the application menu
//  WM_PAINT	- Paint the main window
//  WM_DESTROY	- post a quit message and return
//

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message) 
	{
	case WM_COMMAND:
		wmId    = LOWORD(wParam); 
		wmEvent = HIWORD(wParam); 
		// Parse the menu selections:
		switch (wmId)
		{
		case IDM_ABOUT:
			DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
			break;
		case IDM_EXIT:
			DestroyWindow(hWnd);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);

		EndPaint(hWnd, &ps);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

// Message handler for about box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
		{
			EndDialog(hDlg, LOWORD(wParam));
			return TRUE;
		}
		break;
	}
	return FALSE;
}
